import grpc

from rx.client.configuration import config_base


def get_channel(addr: str) -> grpc.Channel:
  return (
    grpc.insecure_channel(addr) if config_base.is_local() else
    grpc.secure_channel(addr, credentials=grpc.ssl_channel_credentials()))



from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
INVALID: StatusCode
NOT_FOUND: StatusCode
OK: StatusCode
UNAUTHORIZED: StatusCode
UNKNOWN: StatusCode

class AllocNewWorkerReq(_message.Message):
    __slots__ = ["workspace_id"]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ...) -> None: ...

class Delta(_message.Message):
    __slots__ = ["add_dir", "add_path", "remove_path"]
    ADD_DIR_FIELD_NUMBER: _ClassVar[int]
    ADD_PATH_FIELD_NUMBER: _ClassVar[int]
    REMOVE_PATH_FIELD_NUMBER: _ClassVar[int]
    add_dir: _containers.RepeatedScalarFieldContainer[str]
    add_path: _containers.RepeatedScalarFieldContainer[str]
    remove_path: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, add_path: _Optional[_Iterable[str]] = ..., add_dir: _Optional[_Iterable[str]] = ..., remove_path: _Optional[_Iterable[str]] = ...) -> None: ...

class Environment(_message.Message):
    __slots__ = ["alloc", "python"]
    ALLOC_FIELD_NUMBER: _ClassVar[int]
    PYTHON_FIELD_NUMBER: _ClassVar[int]
    alloc: Remote
    python: Python
    def __init__(self, alloc: _Optional[_Union[Remote, _Mapping]] = ..., python: _Optional[_Union[Python, _Mapping]] = ...) -> None: ...

class ExecRequest(_message.Message):
    __slots__ = ["argv", "rsync_source", "workspace_id"]
    ARGV_FIELD_NUMBER: _ClassVar[int]
    RSYNC_SOURCE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    argv: _containers.RepeatedScalarFieldContainer[str]
    rsync_source: RsyncSource
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ..., argv: _Optional[_Iterable[str]] = ..., rsync_source: _Optional[_Union[RsyncSource, _Mapping]] = ...) -> None: ...

class ExecResponse(_message.Message):
    __slots__ = ["execution_id", "output_files", "result", "stderr", "stdout"]
    EXECUTION_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_FILES_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    STDERR_FIELD_NUMBER: _ClassVar[int]
    STDOUT_FIELD_NUMBER: _ClassVar[int]
    execution_id: str
    output_files: _containers.RepeatedScalarFieldContainer[str]
    result: Result
    stderr: bytes
    stdout: bytes
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., execution_id: _Optional[str] = ..., stdout: _Optional[bytes] = ..., stderr: _Optional[bytes] = ..., output_files: _Optional[_Iterable[str]] = ...) -> None: ...

class GetUserResponse(_message.Message):
    __slots__ = ["result", "username"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    result: Result
    username: str
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., username: _Optional[str] = ...) -> None: ...

class InitRequest(_message.Message):
    __slots__ = ["project_name", "rsync_source", "target_env"]
    PROJECT_NAME_FIELD_NUMBER: _ClassVar[int]
    RSYNC_SOURCE_FIELD_NUMBER: _ClassVar[int]
    TARGET_ENV_FIELD_NUMBER: _ClassVar[int]
    project_name: str
    rsync_source: RsyncSource
    target_env: Environment
    def __init__(self, project_name: _Optional[str] = ..., rsync_source: _Optional[_Union[RsyncSource, _Mapping]] = ..., target_env: _Optional[_Union[Environment, _Mapping]] = ...) -> None: ...

class InitResponse(_message.Message):
    __slots__ = ["result", "rsync_dest", "worker_addr", "workspace_id"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    RSYNC_DEST_FIELD_NUMBER: _ClassVar[int]
    WORKER_ADDR_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    result: Result
    rsync_dest: RsyncDestination
    worker_addr: str
    workspace_id: str
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., rsync_dest: _Optional[_Union[RsyncDestination, _Mapping]] = ..., worker_addr: _Optional[str] = ..., workspace_id: _Optional[str] = ...) -> None: ...

class InstallDepsRequest(_message.Message):
    __slots__ = ["workspace_id"]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ...) -> None: ...

class InstallDepsResponse(_message.Message):
    __slots__ = ["result", "stdout"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    STDOUT_FIELD_NUMBER: _ClassVar[int]
    result: Result
    stdout: bytes
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., stdout: _Optional[bytes] = ...) -> None: ...

class KillRequest(_message.Message):
    __slots__ = ["execution_id", "workspace_id"]
    EXECUTION_ID_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    execution_id: str
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ..., execution_id: _Optional[str] = ...) -> None: ...

class Python(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class Remote(_message.Message):
    __slots__ = ["hardware", "image"]
    class Hardware(_message.Message):
        __slots__ = ["processor"]
        PROCESSOR_FIELD_NUMBER: _ClassVar[int]
        processor: str
        def __init__(self, processor: _Optional[str] = ...) -> None: ...
    class Image(_message.Message):
        __slots__ = ["docker", "tag"]
        DOCKER_FIELD_NUMBER: _ClassVar[int]
        TAG_FIELD_NUMBER: _ClassVar[int]
        docker: str
        tag: str
        def __init__(self, docker: _Optional[str] = ..., tag: _Optional[str] = ...) -> None: ...
    HARDWARE_FIELD_NUMBER: _ClassVar[int]
    IMAGE_FIELD_NUMBER: _ClassVar[int]
    hardware: Remote.Hardware
    image: Remote.Image
    def __init__(self, image: _Optional[_Union[Remote.Image, _Mapping]] = ..., hardware: _Optional[_Union[Remote.Hardware, _Mapping]] = ...) -> None: ...

class Result(_message.Message):
    __slots__ = ["code", "message"]
    CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    code: StatusCode
    message: str
    def __init__(self, code: _Optional[_Union[StatusCode, str]] = ..., message: _Optional[str] = ...) -> None: ...

class RsyncDestination(_message.Message):
    __slots__ = ["daemon_module"]
    DAEMON_MODULE_FIELD_NUMBER: _ClassVar[int]
    daemon_module: str
    def __init__(self, daemon_module: _Optional[str] = ...) -> None: ...

class RsyncSource(_message.Message):
    __slots__ = ["directory", "machine_id", "public_key"]
    DIRECTORY_FIELD_NUMBER: _ClassVar[int]
    MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    PUBLIC_KEY_FIELD_NUMBER: _ClassVar[int]
    directory: str
    machine_id: int
    public_key: bytes
    def __init__(self, machine_id: _Optional[int] = ..., directory: _Optional[str] = ..., public_key: _Optional[bytes] = ...) -> None: ...

class SetUsernameRequest(_message.Message):
    __slots__ = ["username"]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    username: str
    def __init__(self, username: _Optional[str] = ...) -> None: ...

class SetUsernameResponse(_message.Message):
    __slots__ = ["result"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    result: Result
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ...) -> None: ...

class StatusCode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import pathlib

from rx.client.configuration import config_base


class Remote(config_base.ReadOnlyConfig):
  def __init__(self, working_dir: pathlib.Path):
    super().__init__(_get_remote_config_file(working_dir))

  @property
  def workspace_id(self):
    return self['workspace_id']

  @property
  def worker_addr(self):
    return self['worker_addr']


class WritableRemote(config_base.ReadWriteConfig):
  def __init__(self, working_dir: pathlib.Path):
    super().__init__(_get_remote_config_file(working_dir))


def _get_remote_config_file(rxroot: pathlib.Path) -> pathlib.Path:
  return config_base.get_config_dir(rxroot) / 'remote'



"""Handle outputs from the server."""
import pathlib
import sys
from typing import List

from rx.client import rsync
from rx.proto import rx_pb2


class OutputHandler:
  """Handle output files and stdout/err."""

  def __init__(self, rxroot: pathlib.Path) -> None:
    self._rxroot = rxroot
    self._current_outputs = set()

  @property
  def remote_paths(self) -> List[str]:
    return sorted([f'{o}' for o in self._current_outputs])

  def handle(self, resp: rx_pb2.ExecResponse):
    """Shows output & error streams and creates output files."""
    if resp.stdout:
      sys.stdout.buffer.write(resp.stdout)
      sys.stdout.flush()
    if resp.stderr:
      sys.stderr.buffer.write(resp.stderr)
      sys.stderr.flush()
    for pth_str in resp.output_files:
      self._current_outputs.add(pathlib.Path(pth_str))

  def write_outputs(self, rsync_client: rsync.RsyncClient):
    if not self._current_outputs:
      return
    rsync_client.from_remote('rx-out', self._rxroot / pathlib.Path('rx-out'))
    print('Created outputs:')
    for pth in self.remote_paths:
      print(f'  {pth}')



"""All of the info we can get without going to the server."""
import hashlib
import json
import pathlib
import shutil
import subprocess
from typing import Optional, Tuple
import uuid

from absl import flags
from absl import logging
import sty

from rx.client.configuration import config_base
from rx.proto import rx_pb2

VERSION = '0.0.6'

IGNORE = pathlib.Path('.rxignore')

_REMOTE = flags.DEFINE_string(
  'remote', None,
  'The path to the remote configuration file to use (see .rx/README.md).')
_RSYNC_PATH = flags.DEFINE_string('rsync_path', None, 'Path to rsync binary')

_REMOTE_DIR = config_base.RX_DIR / 'remotes'
_DEFAULT_REMOTE = _REMOTE_DIR / 'default'


class LocalConfigWriter(config_base.ReadWriteConfig):
  """This holds all of the configuration options that can be determined from
  the local machine."""

  def __init__(self, workspace_dir: pathlib.Path):
    super().__init__(_get_local_config_file(workspace_dir))
    self._workspace_dir = workspace_dir

  def setup_remote(self):
    """Sets the "remote" field (and color) for the local config."""
    remote = str(_REMOTE.value if _REMOTE.value else _DEFAULT_REMOTE)
    remote_path = self._workspace_dir / remote
    try:
      with remote_path.open(mode='rt', encoding='utf-8') as fh:
        remote_content = fh.read()
        self._set_color(remote_content)
        self['remote'] = remote
    except FileExistsError:
      raise FileExistsError(
        f'Could not find remote config: {remote_path} does not exist')

  def _set_color(self, s: str):
    """Generate a random color based on the remote config."""
    h = hashlib.blake2s(digest_size=3)
    h.update(s.encode())
    hash_val = int(h.hexdigest(), 16)
    r = hash_val & 255
    hash_val = hash_val >> 8
    g = hash_val & 255
    hash_val = hash_val >> 8
    b = hash_val & 255
    self['color'] = {'r': r, 'g': g, 'b': b}


class LocalConfig(config_base.ReadOnlyConfig):
  """Create the local configuration."""

  def __init__(self, working_dir: pathlib.Path):
    super().__init__(_get_local_config_file(working_dir))
    self._cwd = working_dir
    self.config_dir = self.cwd / config_base.RX_DIR
    self._color = None

  @property
  def cwd(self) -> pathlib.Path:
    return self._cwd

  @property
  def rsync_source(self) -> rx_pb2.RsyncSource:
    return rx_pb2.RsyncSource(
      machine_id=uuid.getnode(),
      directory=str(self.cwd),
    )

  def _get_source_env(self, docker_image: str) -> rx_pb2.Environment:
    # TODO: how to automatically determine language?
    if 'python' not in docker_image:
      logging.info('Using a non-Python image. There be dragons here.')
      return rx_pb2.Environment()
    return rx_pb2.Environment(python=rx_pb2.Python())

  def get_target_env(self) -> rx_pb2.Environment:
    remote_file = self.cwd / self['remote']
    try:
      with remote_file.open(mode='rt', encoding='utf-8') as fh:
        remote_config = json.load(fh)
    except json.JSONDecodeError as e:
      logging.exception(f'Could not parse {self["remote"]}')
      raise e
    env = rx_pb2.Environment(alloc=rx_pb2.Remote())
    if 'image' not in remote_config or 'docker' not in remote_config['image']:
      raise ConfigError(
        f'Remote config {self["remote"]} must contain "image": '
        '{"docker": "<docker image>"}')
    docker_image = remote_config['image']['docker']
    env.alloc.image.CopyFrom(rx_pb2.Remote.Image(
      docker=docker_image,
    ))
    if 'hardware' in remote_config and 'processor' in remote_config['hardware']:
      env.alloc.hardware.CopyFrom(rx_pb2.Remote.Hardware(
        processor=remote_config['hardware']['processor']
      ))
    env.MergeFrom(self._get_source_env(docker_image))
    return env

  def color_str(self, s: str) -> str:
    color = self['color']
    fg = sty.fg(color['r'], color['g'], color['b'])
    return f'{fg}{s}{sty.rs.fg}'


def create_local_config(cwd: pathlib.Path) -> LocalConfig:
  """Gets or creates .rx directory."""
  config_dir = _get_local_config_file(cwd).parent
  config_dir.mkdir(exist_ok=True, parents=True)
  with LocalConfigWriter(cwd) as c:
    c.setup_remote()
    c['project_name'] = _find_project_name(cwd)
    c['rsync_path'] = _get_rsync_path()
  return LocalConfig(cwd)


def find_local_config(working_dir: pathlib.Path) -> Optional[LocalConfig]:
  """Factory to create a config by looking for .rx."""
  cfg_path = config_base.get_config_dir(working_dir)
  for parent in cfg_path.parents:
    if config_base.get_config_dir(parent).exists():
      return LocalConfig(parent)
  return None


def get_source_path() -> pathlib.Path:
  """Gets the path bundled client files can be found on."""
  # __file__ is ./rx/client/configuration/local.py, so this resolves to ./.
  return pathlib.Path(__file__).resolve().parent.parent.parent.parent


def get_grpc_metadata() -> Tuple[Tuple[str, str]]:
  return (('cv', VERSION),)


def install_local_files(cwd: pathlib.Path):
  # Output directory.
  (cwd / 'rx-out').mkdir(exist_ok=True)

  install_dir = pathlib.Path(get_source_path() / 'install')

  _install_file(install_dir, cwd, IGNORE)

  config_dir = cwd / config_base.RX_DIR
  config_dir.mkdir(exist_ok=True, parents=True)
  _install_file(install_dir, config_dir, 'README.md')

  remotes = cwd / _REMOTE_DIR
  remotes.mkdir(exist_ok=True, parents=True)

  # Copy built-in configs.
  _install_file(install_dir, remotes, 'python-cpu')
  _install_file(install_dir, remotes, 'python-gpu')
  # Create soft link.
  default_config = remotes / 'default'
  if default_config.exists():
    # Don't undo someone else's config.
    return
  # .rx/remote/default -> python-cpu
  default_config.symlink_to('python-cpu')


def _get_rsync_path() -> str:
  """Make sure rsync is installed."""
  if _RSYNC_PATH.value:
    return _RSYNC_PATH.value
  try:
    result = subprocess.run(
      ['which', 'rsync'], check=True, capture_output=True)
  except subprocess.CalledProcessError as e:
    logging.error('Error finding rsync: %s', e)
    raise ConfigError('Cannot find rsync, is it installed/on your path?')
  # Use rsync on PATH
  return result.stdout.decode('utf-8').strip('\n')


def _find_project_name(start_dir: pathlib.Path) -> str:
  """Heuristic to find a reasonable name for this project."""
  # Probably the git repo is rooted on a good name.
  if (start_dir / '.git').exists():
    return start_dir.name
  for parent in start_dir.parents:
    if (parent / '.git').exists():
      return start_dir.name
  # Maybe we haven't initialized git yet, use out name.
  return start_dir.name


def _get_local_config_file(rxroot: pathlib.Path) -> pathlib.Path:
  """Return .rx/trex-dev.run-rx.com/config/local."""
  return config_base.get_config_dir(rxroot) / 'local'


def _install_file(install_dir, config_dir, base_name):
  """Installs a file if it doesn't already exist."""
  source_path = install_dir / base_name
  destination_path = config_dir / base_name
  if not destination_path.exists():
    shutil.copy(source_path, destination_path)


class ConfigError(RuntimeError):
  pass



from collections import abc
import pathlib
import subprocess
import tempfile
from typing import List

from absl import logging

from rx.client.configuration import config_base
from rx.client.configuration import local

def dry_run(local_cfg: local.LocalConfig):
  dest = tempfile.mkdtemp()
  cmd = [
      local_cfg['rsync_path'],
      '--archive',
      '--compress',
      '--delete',
      '--dry-run',
      f'--exclude-from={local_cfg.cwd / local.IGNORE}',
      '--itemize-changes',
      f'{local_cfg.cwd}/',
      str(dest)]
  result = subprocess.run(cmd, check=True, capture_output=True)
  stdout = result.stdout.decode('utf-8')
  print('Uploading:')
  for ln in stdout.split('\n'):
    delta = ln.split(' ')
    if len(delta) < 2 or delta[1] == './':
      continue
    filename = delta[1]
    tab_count = 1 + filename.count('/')
    is_dir = False
    if filename.endswith('/'):
      is_dir = True
      tab_count -= 1
    tabs = '  ' * tab_count
    slash = '/' if is_dir else ''
    pretty_name = f'{pathlib.Path(filename).name}{slash}'
    # TODO: handle file deletion.
    print(f'{tabs}{pretty_name}')

class RsyncClient:
  """Rsync tools."""

  def __init__(
      self, local_cfg: local.LocalConfig, remote_cfg: abc.Mapping):
    self._sync_dir = local_cfg.cwd
    self._cfg = remote_cfg
    self._rsync_path = local_cfg['rsync_path']
    self._daemon_addr = self._cfg["worker_addr"]
    if config_base.is_local():
      # Remove the port (rsync isn't listening on 50051).
      self._daemon_addr = self._daemon_addr.split(':')[0]

  @property
  def host(self) -> str:
    return self._daemon_addr

  @property
  def workspace_id(self) -> str:
    return self._cfg['workspace_id']

  @property
  def _upload_path(self) -> pathlib.Path:
    return pathlib.Path(self._cfg['daemon_module'])

  def from_remote(self, source: str, dest: pathlib.Path) -> int:
    """Copies output files from the remote machine to dest."""
    assert dest.is_dir(), f'Destination {dest} must be a directory'
    assert dest.is_absolute(), f'Destination {dest} must be absolute'
    remote_path = self._upload_path / source
    daemon = f'{self._daemon_addr}::{remote_path}/'
    cmd = [
        self._rsync_path,
        '--archive',
        '--compress',
        '--delete',
        '--quiet',
        f'--exclude-from={self._sync_dir / local.IGNORE}',
        daemon,
        str(dest)]
    return _run_rsync(cmd)

  def to_remote(self) -> int:
    """Copies files/dirs to remote."""
    daemon = f'{self._daemon_addr}::{self._upload_path}'
    cmd = [
        self._rsync_path,
        '--archive',
        '--compress',
        '--delete',
        '--inplace',
        f'--exclude-from={self._sync_dir / local.IGNORE}',
        f'{self._sync_dir}/',
        daemon
    ]
    return _run_rsync(cmd)


def _run_rsync(cmd: List[str]) -> int:
  logging.info('Running %s', cmd)
  try:
    result = subprocess.run(cmd, check=True, capture_output=True)
  except subprocess.CalledProcessError as e:
    logging.error('Error running `%s`', ' '.join(e.cmd))
    if e.returncode == 10:
      # Worker was unreachable.
      logging.error('stderr: %s', e.stderr)
    if e.returncode is None:
      return -1
    return e.returncode
  if result.stdout:
    print(f'stdout: {result.stdout.decode("utf-8")}')
  return 0



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rx/proto/rx.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11rx/proto/rx.proto\x12\x02rx\x1a\x1bgoogle/protobuf/empty.proto\"\x95\x01\n\x06Remote\x12\x1f\n\x05image\x18\x01 \x01(\x0b\x32\x10.rx.Remote.Image\x12%\n\x08hardware\x18\x02 \x01(\x0b\x32\x13.rx.Remote.Hardware\x1a$\n\x05Image\x12\x0e\n\x06\x64ocker\x18\x01 \x01(\t\x12\x0b\n\x03tag\x18\x02 \x01(\t\x1a\x1d\n\x08Hardware\x12\x11\n\tprocessor\x18\x01 \x01(\t\"\x08\n\x06Python\"D\n\x0b\x45nvironment\x12\x19\n\x05\x61lloc\x18\x01 \x01(\x0b\x32\n.rx.Remote\x12\x1a\n\x06python\x18\x02 \x01(\x0b\x32\n.rx.Python\"?\n\x05\x44\x65lta\x12\x10\n\x08\x61\x64\x64_path\x18\x01 \x03(\t\x12\x0f\n\x07\x61\x64\x64_dir\x18\x02 \x03(\t\x12\x13\n\x0bremove_path\x18\x03 \x03(\t\"7\n\x06Result\x12\x1c\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x0e.rx.StatusCode\x12\x0f\n\x07message\x18\x02 \x01(\t\"H\n\x0bRsyncSource\x12\x12\n\nmachine_id\x18\x01 \x01(\x03\x12\x11\n\tdirectory\x18\x02 \x01(\t\x12\x12\n\npublic_key\x18\x03 \x01(\x0c\")\n\x10RsyncDestination\x12\x15\n\rdaemon_module\x18\x01 \x01(\t\")\n\x11\x41llocNewWorkerReq\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\"X\n\x0b\x45xecRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgv\x18\x02 \x03(\t\x12%\n\x0crsync_source\x18\x03 \x01(\x0b\x32\x0f.rx.RsyncSource\"v\n\x0c\x45xecResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\x14\n\x0c\x65xecution_id\x18\x02 \x01(\t\x12\x0e\n\x06stdout\x18\x03 \x01(\x0c\x12\x0e\n\x06stderr\x18\x04 \x01(\x0c\x12\x14\n\x0coutput_files\x18\x05 \x03(\t\"?\n\x0fGetUserResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\x10\n\x08username\x18\x02 \x01(\t\"o\n\x0bInitRequest\x12\x14\n\x0cproject_name\x18\x01 \x01(\t\x12%\n\x0crsync_source\x18\x02 \x01(\x0b\x32\x0f.rx.RsyncSource\x12#\n\ntarget_env\x18\x03 \x01(\x0b\x32\x0f.rx.Environment\"\x7f\n\x0cInitResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12(\n\nrsync_dest\x18\x02 \x01(\x0b\x32\x14.rx.RsyncDestination\x12\x13\n\x0bworker_addr\x18\x03 \x01(\t\x12\x14\n\x0cworkspace_id\x18\x04 \x01(\t\"*\n\x12InstallDepsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\"A\n\x13InstallDepsResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\x0e\n\x06stdout\x18\x02 \x01(\x0c\"9\n\x0bKillRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x14\n\x0c\x65xecution_id\x18\x02 \x01(\t\"&\n\x12SetUsernameRequest\x12\x10\n\x08username\x18\x01 \x01(\t\"1\n\x13SetUsernameResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result*O\n\nStatusCode\x12\x06\n\x02OK\x10\x00\x12\x0b\n\x07UNKNOWN\x10\x01\x12\x0b\n\x07INVALID\x10\x02\x12\r\n\tNOT_FOUND\x10\x03\x12\x10\n\x0cUNAUTHORIZED\x10\x04\x32\xb8\x01\n\x10\x45xecutionService\x12\x42\n\x0bInstallDeps\x12\x16.rx.InstallDepsRequest\x1a\x17.rx.InstallDepsResponse\"\x00\x30\x01\x12-\n\x04\x45xec\x12\x0f.rx.ExecRequest\x1a\x10.rx.ExecResponse\"\x00\x30\x01\x12\x31\n\x04Kill\x12\x0f.rx.KillRequest\x1a\x16.google.protobuf.Empty\"\x00\x32\xf4\x01\n\x0cSetupService\x12;\n\x0e\x41llocNewWorker\x12\x15.rx.AllocNewWorkerReq\x1a\x10.rx.InitResponse\"\x00\x12\x38\n\x07GetUser\x12\x16.google.protobuf.Empty\x1a\x13.rx.GetUserResponse\"\x00\x12+\n\x04Init\x12\x0f.rx.InitRequest\x1a\x10.rx.InitResponse\"\x00\x12@\n\x0bSetUsername\x12\x16.rx.SetUsernameRequest\x1a\x17.rx.SetUsernameResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rx.proto.rx_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATUSCODE._serialized_start=1346
  _STATUSCODE._serialized_end=1425
  _REMOTE._serialized_start=55
  _REMOTE._serialized_end=204
  _REMOTE_IMAGE._serialized_start=137
  _REMOTE_IMAGE._serialized_end=173
  _REMOTE_HARDWARE._serialized_start=175
  _REMOTE_HARDWARE._serialized_end=204
  _PYTHON._serialized_start=206
  _PYTHON._serialized_end=214
  _ENVIRONMENT._serialized_start=216
  _ENVIRONMENT._serialized_end=284
  _DELTA._serialized_start=286
  _DELTA._serialized_end=349
  _RESULT._serialized_start=351
  _RESULT._serialized_end=406
  _RSYNCSOURCE._serialized_start=408
  _RSYNCSOURCE._serialized_end=480
  _RSYNCDESTINATION._serialized_start=482
  _RSYNCDESTINATION._serialized_end=523
  _ALLOCNEWWORKERREQ._serialized_start=525
  _ALLOCNEWWORKERREQ._serialized_end=566
  _EXECREQUEST._serialized_start=568
  _EXECREQUEST._serialized_end=656
  _EXECRESPONSE._serialized_start=658
  _EXECRESPONSE._serialized_end=776
  _GETUSERRESPONSE._serialized_start=778
  _GETUSERRESPONSE._serialized_end=841
  _INITREQUEST._serialized_start=843
  _INITREQUEST._serialized_end=954
  _INITRESPONSE._serialized_start=956
  _INITRESPONSE._serialized_end=1083
  _INSTALLDEPSREQUEST._serialized_start=1085
  _INSTALLDEPSREQUEST._serialized_end=1127
  _INSTALLDEPSRESPONSE._serialized_start=1129
  _INSTALLDEPSRESPONSE._serialized_end=1194
  _KILLREQUEST._serialized_start=1196
  _KILLREQUEST._serialized_end=1253
  _SETUSERNAMEREQUEST._serialized_start=1255
  _SETUSERNAMEREQUEST._serialized_end=1293
  _SETUSERNAMERESPONSE._serialized_start=1295
  _SETUSERNAMERESPONSE._serialized_end=1344
  _EXECUTIONSERVICE._serialized_start=1428
  _EXECUTIONSERVICE._serialized_end=1612
  _SETUPSERVICE._serialized_start=1615
  _SETUPSERVICE._serialized_end=1859
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from rx.proto import rx_pb2 as rx_dot_proto_dot_rx__pb2


class ExecutionServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.InstallDeps = channel.unary_stream(
                '/rx.ExecutionService/InstallDeps',
                request_serializer=rx_dot_proto_dot_rx__pb2.InstallDepsRequest.SerializeToString,
                response_deserializer=rx_dot_proto_dot_rx__pb2.InstallDepsResponse.FromString,
                )
        self.Exec = channel.unary_stream(
                '/rx.ExecutionService/Exec',
                request_serializer=rx_dot_proto_dot_rx__pb2.ExecRequest.SerializeToString,
                response_deserializer=rx_dot_proto_dot_rx__pb2.ExecResponse.FromString,
                )
        self.Kill = channel.unary_unary(
                '/rx.ExecutionService/Kill',
                request_serializer=rx_dot_proto_dot_rx__pb2.KillRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class ExecutionServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def InstallDeps(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Exec(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Kill(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ExecutionServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'InstallDeps': grpc.unary_stream_rpc_method_handler(
                    servicer.InstallDeps,
                    request_deserializer=rx_dot_proto_dot_rx__pb2.InstallDepsRequest.FromString,
                    response_serializer=rx_dot_proto_dot_rx__pb2.InstallDepsResponse.SerializeToString,
            ),
            'Exec': grpc.unary_stream_rpc_method_handler(
                    servicer.Exec,
                    request_deserializer=rx_dot_proto_dot_rx__pb2.ExecRequest.FromString,
                    response_serializer=rx_dot_proto_dot_rx__pb2.ExecResponse.SerializeToString,
            ),
            'Kill': grpc.unary_unary_rpc_method_handler(
                    servicer.Kill,
                    request_deserializer=rx_dot_proto_dot_rx__pb2.KillRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'rx.ExecutionService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ExecutionService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def InstallDeps(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/rx.ExecutionService/InstallDeps',
            rx_dot_proto_dot_rx__pb2.InstallDepsRequest.SerializeToString,
            rx_dot_proto_dot_rx__pb2.InstallDepsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Exec(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/rx.ExecutionService/Exec',
            rx_dot_proto_dot_rx__pb2.ExecRequest.SerializeToString,
            rx_dot_proto_dot_rx__pb2.ExecResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Kill(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/rx.ExecutionService/Kill',
            rx_dot_proto_dot_rx__pb2.KillRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class SetupServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AllocNewWorker = channel.unary_unary(
                '/rx.SetupService/AllocNewWorker',
                request_serializer=rx_dot_proto_dot_rx__pb2.AllocNewWorkerReq.SerializeToString,
                response_deserializer=rx_dot_proto_dot_rx__pb2.InitResponse.FromString,
                )
        self.GetUser = channel.unary_unary(
                '/rx.SetupService/GetUser',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=rx_dot_proto_dot_rx__pb2.GetUserResponse.FromString,
                )
        self.Init = channel.unary_unary(
                '/rx.SetupService/Init',
                request_serializer=rx_dot_proto_dot_rx__pb2.InitRequest.SerializeToString,
                response_deserializer=rx_dot_proto_dot_rx__pb2.InitResponse.FromString,
                )
        self.SetUsername = channel.unary_unary(
                '/rx.SetupService/SetUsername',
                request_serializer=rx_dot_proto_dot_rx__pb2.SetUsernameRequest.SerializeToString,
                response_deserializer=rx_dot_proto_dot_rx__pb2.SetUsernameResponse.FromString,
                )


class SetupServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AllocNewWorker(self, request, context):
        """Creates a new worker for an existing workspace.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetUser(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Init(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetUsername(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SetupServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AllocNewWorker': grpc.unary_unary_rpc_method_handler(
                    servicer.AllocNewWorker,
                    request_deserializer=rx_dot_proto_dot_rx__pb2.AllocNewWorkerReq.FromString,
                    response_serializer=rx_dot_proto_dot_rx__pb2.InitResponse.SerializeToString,
            ),
            'GetUser': grpc.unary_unary_rpc_method_handler(
                    servicer.GetUser,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=rx_dot_proto_dot_rx__pb2.GetUserResponse.SerializeToString,
            ),
            'Init': grpc.unary_unary_rpc_method_handler(
                    servicer.Init,
                    request_deserializer=rx_dot_proto_dot_rx__pb2.InitRequest.FromString,
                    response_serializer=rx_dot_proto_dot_rx__pb2.InitResponse.SerializeToString,
            ),
            'SetUsername': grpc.unary_unary_rpc_method_handler(
                    servicer.SetUsername,
                    request_deserializer=rx_dot_proto_dot_rx__pb2.SetUsernameRequest.FromString,
                    response_serializer=rx_dot_proto_dot_rx__pb2.SetUsernameResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'rx.SetupService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SetupService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AllocNewWorker(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/rx.SetupService/AllocNewWorker',
            rx_dot_proto_dot_rx__pb2.AllocNewWorkerReq.SerializeToString,
            rx_dot_proto_dot_rx__pb2.InitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/rx.SetupService/GetUser',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            rx_dot_proto_dot_rx__pb2.GetUserResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Init(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/rx.SetupService/Init',
            rx_dot_proto_dot_rx__pb2.InitRequest.SerializeToString,
            rx_dot_proto_dot_rx__pb2.InitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetUsername(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/rx.SetupService/SetUsername',
            rx_dot_proto_dot_rx__pb2.SetUsernameRequest.SerializeToString,
            rx_dot_proto_dot_rx__pb2.SetUsernameResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from collections import abc
import json
import pathlib
from types import TracebackType
from typing import Optional, Type

from absl import flags

RX_DIR = pathlib.Path('.rx')

# Configuration files are placed under a directory with this name (this is just
# to make development easier, so we can have separate configs for
# local/dev/prod).
TREX_HOST = flags.DEFINE_string(
  'trex-host', 'trex.run-rx.com', 'GRPC host to connect to.')
RX_ROOT = flags.DEFINE_string(
  'rxroot', None, 'The directory to use as rxroot (defaults to pwd)')


class ReadOnlyConfig(abc.Mapping):
  def __init__(self, config_file: pathlib.Path, strict_mode: bool = True):
    self._config_file = config_file
    if not strict_mode and not config_file.exists():
      self._config = {}
      return
    with self._config_file.open(mode='rt', encoding='utf-8') as fh:
      self._config = json.load(fh)

  def __getitem__(self, item):
    assert item in self._config, (
      f'Could not access key {item} in {self._config_file}')
    return self._config[item]

  def __iter__(self):
    return self._config.__iter__()

  def __len__(self) -> int:
    return len(self._config)


class ReadWriteConfig(abc.MutableMapping, ReadOnlyConfig):

  def __init__(self, config_file: pathlib.Path):
    super().__init__(config_file, strict_mode=False)

  def __enter__(self):
    return self

  def __exit__(self, exctype: Optional[Type[BaseException]],
             excinst: Optional[BaseException],
             exctb: Optional[TracebackType]) -> bool:
    if exctype is None:
      # Keep config file up-to-date.
      with self._config_file.open(mode='wt', encoding='utf-8') as fh:
        json.dump(self._config, fh)
    return False

  def __setitem__(self, item, value):
    self._config[item] = value

  def __delitem__(self, item):
    self._config.__delitem__(item)


def get_config_dir(rxroot: pathlib.Path) -> pathlib.Path:
  """Returns the absolute config directory path, e.g., /proj/.rx/t.c/config."""
  return rxroot / RX_DIR / TREX_HOST.value / 'config'


def is_local() -> bool:
  return TREX_HOST.value.startswith('localhost')



"""Handles client-side user functionality.

The user info is stored in .rx/user/config and includes the user's username.
"""
import pathlib
import re
from typing import Optional

from rx.client import menu
from rx.client.configuration import config_base

_VALID_USERNAME = re.compile(r'^[a-zA-Z][\w]{2,14}$')


class CreateUser(config_base.ReadWriteConfig):
  def __init__(self, cwd: pathlib.Path):
    super().__init__(_get_user_config(cwd))


class User(config_base.ReadOnlyConfig):
  """User info."""
  def __init__(self, cwd: pathlib.Path, email: str):
    super().__init__(_get_user_config(cwd))
    if email != self['email']:
      raise RuntimeError(f'Mismatched emails: {email} vs. {self["email"]}')


def has_config(cwd: pathlib.Path) -> bool:
  return _get_user_config(cwd).exists()


def username_prompt(email: str) -> str:
  """Prompts the user to choose a username."""
  return menu.string_prompt(
      f'Choose a username for {email}', validation=username_validator)


def username_validator(username: Optional[str]) -> bool:
  """Returns if username is valid."""
  is_valid = (
    username is not None and _VALID_USERNAME.match(username) is not None)
  if not is_valid:
    if username is None or len(username) < 3:
      print('Error: username too short.')
    elif username is not None and len(username) > 15:
      print('Error: username too long.')
    else:
      print('Error: username must contain only alphanumeric characters.')
  return is_valid


def _get_user_config(rxroot: pathlib.Path) -> pathlib.Path:
  return rxroot / config_base.get_config_dir(rxroot) / 'user/config'



"""Text menus.

Each helper function has the same general structure:

1. Setup the prompt.
2. Convert the response to the expected type.
2. Validate the response. If validation fails, go back to 1.
"""

import pathlib
from typing import Any, Callable, List, Optional


def bool_prompt(prompt: str) -> bool:
  response = input(f'{prompt} (y/n): ')
  if response == 'y':
    return True
  elif response == 'n':
    return False
  return bool_prompt(prompt)


def numbered_options_prompt(
    prompt: str, options: List[Any], default: int = 0) -> int:
  """Prompt with a numbered list of options."""

  # Setup
  assert options
  print(prompt)
  for i, p in options:
    print(f'{i}: {p}')
  response = input(f'[{default}]: ')

  # Convert.
  if not response:
    return 0
  response = int(response)

  # Validation.
  if response >= len(options):
    print(f'Must be between 0 and {len(options) - 1}')
    return numbered_options_prompt(prompt, options, default)
  return response


def path_prompt(
  prompt: str,
  default_path: Optional[pathlib.Path],
  validation: Callable[[Optional[pathlib.Path]], bool] = lambda x: True
) -> pathlib.Path:
  """Prompt for filesystem path."""

  # Setup.
  if default_path:
    response = input(f'{prompt}\n[{default_path}]: ')
  else:
    response = input(f'{prompt}: ')

  # Convert.
  if not response and default_path:
    response = default_path
  else:
    response = pathlib.Path(response)

  # Validation.
  if not validation(response):
    return path_prompt(prompt, default_path, validation)
  return response


def string_prompt(
  prompt: str,
  default_str: Optional[str] = None,
  validation:  Callable[[Optional[str]], bool] = lambda x: True
) -> str:
  # Setup.
  if default_str:
    response = input(f'{prompt}\n[{default_str}]: ')
  else:
    response = input(f'{prompt}: ')

  # Convert.
  if not response and default_str:
    response = default_str

  # Validation.
  if not validation(response):
    return string_prompt(prompt, default_str, validation)
  return response



import sys
from typing import cast

from absl import logging
from google.protobuf import empty_pb2
import grpc

from rx.client import grpc_helper
from rx.client import login
from rx.client import rsync
from rx.client import user
from rx.client.configuration import local
from rx.client.configuration import remote
from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc

_TIMEOUT = 10

class Client():
  """Handle contacting the remote server."""

  def __init__(self, channel: grpc.Channel, local_cfg: local.LocalConfig):
    self._stub = rx_pb2_grpc.SetupServiceStub(channel)
    self._local_cfg = local_cfg
    self._login = login.LoginManager(local_cfg.cwd)
    self._metadata = local.get_grpc_metadata()

  def create_user_or_log_in(self) -> user.User:
    # First, make sure we're logged in with Google.
    try:
      self._login.login()
    except login.AuthError as e:
      raise InitError(e, -1)
    self._metadata += self._login.grpc_metadata
    email = self._login.id_token['email']

    # Check if user is set up.
    # TODO: it would be better to read the username from the config and
    # automatically try to set it on the remote. However, this generally
    # shouldn't come up for normal users.
    username = self._get_username()
    if not user.has_config(self._local_cfg.cwd):
      with user.CreateUser(self._local_cfg.cwd) as c:
        c['username'] = username
        c['email'] = email
    return user.User(self._local_cfg.cwd, email)

  def dry_run(self):
    rsync.dry_run(self._local_cfg)

  def init(self) -> int:
    try:
      target_env = self._local_cfg.get_target_env()
    except local.ConfigError as e:
      raise InitError(str(e), -1)
    if target_env.alloc.hardware.processor == 'gpu':
      print('rx only works with CPUs at the moment, but I appreciate your '
            'enthusiasm! Try setting --remote=python-cpu for now.')
      return -1
    req = rx_pb2.InitRequest(
      project_name=self._local_cfg['project_name'],
      rsync_source=self._local_cfg.rsync_source,
      target_env=target_env,
    )
    sys.stdout.write('Finding a remote worker... ')
    sys.stdout.flush()
    try:
      # Five minute timeout.
      resp = self._stub.Init(req, metadata=self._metadata, timeout=(5 * 60))
      sys.stdout.write('Done.\n')
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise InitError(f'Could not initialize worker: {e.details()}', -1)
    if resp.result.code != 0:
      raise InitError(resp.result.message, -1)

    # TODO: create a threaded UserStatus class with __enter__/__exit__.
    sys.stdout.write('Copying source code... ')
    sys.stdout.flush()
    with remote.WritableRemote(self._local_cfg.cwd) as r:
      r['workspace_id'] = resp.workspace_id
      r['worker_addr'] = resp.worker_addr
      r['daemon_module'] = resp.rsync_dest.daemon_module
    self._run_initial_rsync()
    sys.stdout.write('Done.\n')
    self._install_deps(resp.worker_addr, resp.workspace_id)
    print('\nDone setting up rx! To use, run:\n\n\t$ rx <your command>\n')
    return 0

  def _create_username(self) -> str:
    username = user.username_prompt(self._login.id_token['email'])
    req = rx_pb2.SetUsernameRequest(username=username)
    try:
      resp = self._stub.SetUsername(
        req, metadata=self._metadata, timeout=_TIMEOUT)
    except grpc.RpcError as e:
      raise InitError(f'Could not set username: {e.details()}', -1)
    if resp.result.code == rx_pb2.INVALID:
      raise InitError(
        f'{resp.result.message}\nInvalid username: {username}', rx_pb2.INVALID)
    return username

  def _get_username(self) -> str:
    # Check with rx server.
    username = self._get_username_from_rx()
    if username:
      return username

    # Prompt the user to choose a username.
    return self._create_username()

  def _get_username_from_rx(self) -> str:
    try:
      resp = self._stub.GetUser(
        empty_pb2.Empty(), metadata=self._metadata, timeout=_TIMEOUT)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise InitError(f'Could not get user from rx: {e.details()}', -1)
    return resp.username

  def _run_initial_rsync(self):
    r = rsync.RsyncClient(
      self._local_cfg, remote.Remote(self._local_cfg.cwd))
    return_code = r.to_remote()
    if return_code == 0:
      logging.info('Copied files to %s', r.host)

  def _install_deps(self, worker_addr: str, workspace_id: str) -> int:
    channel = grpc_helper.get_channel(worker_addr)
    stub = rx_pb2_grpc.ExecutionServiceStub(channel)
    req = rx_pb2.InstallDepsRequest(workspace_id=workspace_id)
    resp = None
    try:
      for resp in stub.InstallDeps(
        req, metadata=self._metadata, timeout=(10 * 60)):
        if resp.stdout:
          sys.stdout.buffer.write(resp.stdout)
          sys.stdout.buffer.flush()
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise InitError(e.details(), -1)
    if resp and resp.HasField('result') and resp.result.code:
      raise InitError(resp.result.message, resp.result.code)
    return 0


class InitError(RuntimeError):
  """Class to repackage any init errors that happen and add an exit code."""

  def __init__(self, message, code, *args):
    super().__init__(message, *args)
    self._code = code

  @property
  def code(self):
    return self._code



import sys
from typing import List, cast

from absl import logging
import grpc

from rx.client import login
from rx.client import output_handler
from rx.client import rsync
from rx.client.configuration import local
from rx.client.configuration import remote
from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc

SIGINT_CODE = 130

# TODO: set up UI for looking at previous executions.
_BASE_URI = 'https://la-brea.run-rx.com'
_NOT_REACHABLE = 1


class Client():
  """Handle contacting the remote server."""

  def __init__(
      self,
      channel: grpc.Channel,
      local_cfg: local.LocalConfig,
      remote_cfg: remote.Remote):
    self._uri = remote_cfg.worker_addr
    self._rsync = rsync.RsyncClient(local_cfg, remote_cfg)
    self._stub = rx_pb2_grpc.ExecutionServiceStub(channel)
    self._metadata = local.get_grpc_metadata()
    self._local_cfg = local_cfg
    self._remote_cfg = remote_cfg
    self._login = login.LoginManager(local_cfg.cwd)
    self._current_execution_id = None

  def exec(self, argv: List[str]) -> int:
    cmd_str = ' '.join(argv)
    logging.info(f'Running `{cmd_str}` on {self._uri}')

    self._login.login()
    self._metadata += self._login.grpc_metadata

    result = self._rsync.to_remote()
    if result != 0:
      raise UnreachableError(self._uri, result)
    result = None

    request = rx_pb2.ExecRequest(
      workspace_id=self._remote_cfg.workspace_id,
      argv=argv,
      rsync_source=self._local_cfg.rsync_source)

    out_handler = output_handler.OutputHandler(self._local_cfg.cwd)
    response = None
    try:
      for response in self._stub.Exec(request, metadata=self._metadata):
        self._current_execution_id = response.execution_id
        out_handler.handle(response)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      sys.stderr.write(f'Error contacting {self._uri}: {e.details()}\n')
      return _NOT_REACHABLE
    except KeyboardInterrupt:
      self.maybe_kill()
      return SIGINT_CODE
    except RuntimeError as e:
      # Bug in grpc, probably.
      msg = str(e)
      if msg == 'cannot release un-acquired lock':
        self.maybe_kill()
        return SIGINT_CODE
      else:
        sys.stderr.write(msg)
        return -1

    out_handler.write_outputs(self._rsync)

    if (response is not None and response.HasField('result') and
        response.result.code != 0):
      sys.stderr.write(f'{response.result.message}\n')
      return response.result.code
    return 0

  def maybe_kill(self):
    """Kill the process, if it exists."""
    if self._current_execution_id is None:
      logging.error('No ID to kill.')
      return

    req = rx_pb2.KillRequest(
      workspace_id=self._remote_cfg.workspace_id,
      execution_id=self._current_execution_id,
    )
    self._stub.Kill(req, metadata=self._metadata)


class UnreachableError(RuntimeError):
  def __init__(self, worker_addr: str, code: int, *args: object) -> None:
    super().__init__(*args)
    self.worker = worker_addr
    self.code = code



"""Rx usage

Configure this directory to be able to run on a remote host:

    rx init

You only have to run this once per directory, similar to `git init`.

To run a command on a configured remote host:

    rx CMD

"""

from absl import app

from rx.client.commands import exec

app.run(exec.main)



"""Implements client login flow.

See https://developers.google.com/identity/protocols/oauth2/native-app for
details about this flow.

There are two important files:

* .rx/<addr>/user/.refresh-token
* .rx/<addr>/user/access-token

These start out the same. Access tokens expire in 1 hour, so after that, the
'refresh_token' from .refresh-token.json must be used to get a new access token.
However, the response to that request does not contain the refresh token itself,
so it is stored in .rx/user/access-token.json, which should be used for access.

Thus, .refresh-token is something of an implementation detail and access-token
is more "user-facing."
"""
import functools
import http.cookies
import http.server
import json
import pathlib
import platform
import subprocess
import threading
import time
from typing import Any, Dict, Tuple
from urllib import parse

from absl import flags
from absl import logging
import jwt
import requests
from requests import exceptions

from rx.client.configuration import config_base

_DO_AUTH = flags.DEFINE_bool(
  'do_auth', True, 'Skip auth for offline development')
_AUTH_EMAIL = flags.DEFINE_string(
  'auth_email', None, 'Email to use for offline development')

_DARWIN = 'Darwin'
_LINUX = 'Linux'
_WINDOWS = 'Windows'

# From https://console.developers.google.com/apis/credentials
_CLIENT_ID = '909912915303-enmim0gms3hfsdv7c9m6p79vovjkf4vd.apps.googleusercontent.com'
_CLIENT_SECRET = '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'


class Handler(http.server.SimpleHTTPRequestHandler):
  """Handle the response from accounts.google.com."""

  def __init__(self, holder: 'CodeHolder', *args, **kwargs):
    self._holder = holder
    super().__init__(*args, **kwargs)

  def do_GET(self):
    parsed_url = parse.urlparse(self.path)
    code = parse.parse_qs(parsed_url.query)['code'][0]
    self._holder.set_code(code)
    self.send_response(200)
    self.end_headers()
    self.wfile.write(b'All set, feel free to close this tab')


class CodeHolder:
  def __init__(self) -> None:
    self._code_is_set = threading.Event()
    self._code = None

  @property
  def code(self) -> str:
    assert self._code
    return self._code

  def set_code(self, code: str):
    self._code = code
    self._code_is_set.set()

  def wait(self):
    self._code_is_set.wait()


class LoginManager:
  """Handle browser login."""

  def __init__(self, rxroot: pathlib.Path) -> None:
    self._rxroot = rxroot
    self._server = None
    self._port = 0
    self._code_holder = CodeHolder()
    self._access_token = config_base.ReadOnlyConfig(
      _get_access_token_file(rxroot), strict_mode=False)
    if not _DO_AUTH.value:
      self.login = _no_auth_login

  @property
  def grpc_metadata(self) -> Tuple[Tuple[str, str]]:
    """Returns the signed ID token."""
    return (('id-token', self._access_token['id_token']),)

  @property
  def id_token(self) -> Dict[str, str]:
    """Returns the decoded ID token."""
    return decode_id_token(self._access_token['id_token'])

  def login(self) -> config_base.ReadOnlyConfig:
    """Logs in and returns the access token dict.

    Dict is of the form:

    {
      "access_token": "abc...123",
      "expires_in": 3599,
      "scope": "openid https://www.googleapis.com/auth/userinfo.email",
      "token_type": "Bearer",
      "id_token": "<base64-encoded 3-part string>",
    }
    """
    user_dir = config_base.get_config_dir(self._rxroot) / 'user'
    if not user_dir.exists():
      user_dir.mkdir(parents=True, exist_ok=True)
    if not self._access_token:
      self.start_server()
      self.open_browser()
      self.wait_for_login()
    self.validate_login()
    return self._access_token

  def open_browser(self):
    """Opens the browser to the login page."""
    while self._server is None:
      time.sleep(1)
    url = self._create_login_url()
    system = platform.system()
    if system == _DARWIN:
      cmd = ['open', url]
    elif system == _LINUX:
      cmd = ['xdg-open', url]
    elif system == _WINDOWS:
      logging.info(
          'I can\'t imagine this is going to work on Windows, but feel free to '
          'give it a try and report back')
      cmd = ['cmd', '/c', 'start', url.replace('&', '^&')]
    else:
      raise RuntimeError(f'Unsupported system: {system}')
    try:
      subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
      if e.stdout:
        print(e.stdout)
      self._server.shutdown()
      raise AuthError(f'Error opening browser: {e}')
    except FileNotFoundError as e:
      self._server.shutdown()
      raise AuthError(
        f'Could not find browser-opening program {cmd[0]}, are you in an SSH '
        'session?')

  def refresh_access_token(self):
    """Refreshes the access token when it expires."""
    refresh_file = _get_refresh_token_file(self._rxroot)
    if not refresh_file.exists():
      # Access token was expired but the refresh token doesn't exist. Just start
      # over.
      _delete_auth_files(self._rxroot)
      raise AuthError('Could not log in, please try again.')
    with open(refresh_file, mode='rt', encoding='utf-8') as fh:
      refresh_token = json.load(fh)
    data = {
      'client_id': _CLIENT_ID,
      'client_secret': _CLIENT_SECRET,
      'grant_type': 'refresh_token',
      'refresh_token': refresh_token['refresh_token']
    }
    try:
      resp = requests.post(
        'https://oauth2.googleapis.com/token', data=data, timeout=10)
    except exceptions.ConnectionError:
      raise AuthError(
        f'Unable to connect to oauth2.googleapis.com, is your internet up?')
    if resp.status_code != 200:
      # Refreshing went wrong, remove everything!
      _delete_auth_files(self._rxroot)
      raise AuthError(
        f'Unable to refresh log in: {resp.reason} [{resp.status_code}]\n' +
        resp.text)
    access_token = _get_access_token_file(self._rxroot)
    with access_token.open(mode='wt', encoding='utf-8') as fh:
      fh.write(resp.text)
    self._access_token = json.loads(resp.text)

  def start_server(self):
    """Starts a local server in a separate thread."""
    th = threading.Thread(target=self._start_local_server)
    th.start()

  def validate_login(self):
    assert self._access_token
    try:
      id_token = decode_id_token(self._access_token['id_token'])
    except ValueError as e:
      _delete_auth_files(self._rxroot)
      logging.exception(e)
      raise AuthError('Could not read token, please try logging in again.')
    if is_expired(id_token):
      self.refresh_access_token()

  def wait_for_login(self):
    """Shuts down the server after login."""
    self._code_holder.wait()
    assert self._server
    self._server.shutdown()
    self._get_access_token(self._code_holder.code)

  def _create_login_url(self) -> str:
    """Generate the login URL."""
    return (
      'https://accounts.google.com/o/oauth2/v2/auth?'
      'response_type=code&'
      f'client_id={_CLIENT_ID}&'
      'scope=openid%20email&'
      f'redirect_uri=http%3A//localhost:{self._port}')

  def _get_access_token(self, code: str):
    """Exchange the identity code for an access token."""
    data = {
      'client_id': _CLIENT_ID,
      'client_secret': _CLIENT_SECRET,
      'code': code,
      'grant_type': 'authorization_code',
      # We need to provide the matching redirect URI (even though the server
      # isn't even running anymore) or Google will return a 400 error.
      'redirect_uri': f'http://localhost:{self._port}'
    }
    resp = requests.post(
      'https://oauth2.googleapis.com/token', data=data, timeout=10)
    if resp.status_code != 200:
      raise AuthError(
        f'Unable to log in: {resp.reason} [{resp.status_code}]\n{resp.text}')
    access_token = _get_access_token_file(self._rxroot)
    with access_token.open(mode='wt', encoding='utf-8') as fh:
      fh.write(resp.text)
    refresh = _get_refresh_token_file(self._rxroot)
    with refresh.open(mode='wt', encoding='utf-8') as fh:
      fh.write(resp.text)
    self._access_token = json.loads(resp.text)

  def _start_local_server(self):
    """Starts the server."""
    partial_handler = functools.partial(Handler, self._code_holder)
    self._server = http.server.HTTPServer(('localhost', 0), partial_handler)
    self._port = self._server.server_port
    self._server.serve_forever()


def _delete_auth_files(rxroot: pathlib.Path):
  """Remove all files associated with logging in."""
  if _get_refresh_token_file(rxroot).exists():
    _get_refresh_token_file(rxroot).unlink()
  if _get_access_token_file(rxroot).exists():
    _get_access_token_file(rxroot).unlink()


def is_expired(id_token: Dict[str, Any]) -> bool:
  """Returns if the access token is expired."""
  exp = id_token['exp']
  now = time.time()
  return now > exp


def decode_id_token(b64: str) -> Dict[str, Any]:
  """The id_token field is a base64 encoded, .-delimited string."""
  return jwt.decode(b64, options={'verify_signature': False})


def _get_access_token_file(rxroot: pathlib.Path) -> pathlib.Path:
  return config_base.get_config_dir(rxroot) / 'user/access-token'


def _get_refresh_token_file(rxroot: pathlib.Path) -> pathlib.Path:
  return config_base.get_config_dir(rxroot) / 'user/.refresh-token'


def _no_auth_login() -> Dict[str, Any]:
  """Returns an auth token with email: foo@example.com."""
  assert not _DO_AUTH.value and _AUTH_EMAIL.value is not None
  id_token = jwt.encode({'email': _AUTH_EMAIL.value}, 'abc123')
  return {
    'access_token': 'abc123',
    'expires_in': 3599,
    'scope': 'openid https://www.googleapis.com/auth/userinfo.email',
    'token_type': 'Bearer',
    'id_token': id_token,
  }


class AuthError(RuntimeError):
  pass



"""rx usage

To set up this directory on a remote host:

    rx init

You only have to run this once per directory, similar to `git init`.

To run a command on the remote host:

    rx CMD

Run 'rx --help' for more options or visit https://www.run-rx.com.
"""
import pathlib
import tempfile
from typing import List

from absl import app
from absl import logging

from rx.client import executor
from rx.client import grpc_helper
from rx.client.commands import init
from rx.client.configuration import config_base
from rx.client.configuration import local
from rx.client.configuration import remote


class ExecCommand:

  def __init__(self, argv: List[str]):
    self._argv = argv
    cwd = (
      pathlib.Path(config_base.RX_ROOT.value) if config_base.RX_ROOT.value else
      pathlib.Path.cwd())
    self._config = local.find_local_config(cwd)

  def run(self) -> int:
    if self._config is None:
      print ('Run `rx init` first!')
      return -1
    remote_cfg = remote.Remote(self._config.cwd)
    with grpc_helper.get_channel(remote_cfg.worker_addr) as ch:
      client = executor.Client(ch, self._config, remote_cfg)
      return self._try_exec(client)

  def _try_exec(self, client: executor.Client) -> int:
    """Sends the command to the server."""
    if len(self._argv) < 1:
      print('No command given.')
      return 1

    if len(self._argv) == 1:
      # If this is a quoted arg, break it into pieces. E.g., to pass "ls -l" it
      # must be quoted so that absl doesn't try to capture the -l.
      # TODO: is there a better way to handle this with absl flags?
      self._argv = self._argv[0].split(' ')
    try:
      return client.exec(self._argv)
    except KeyboardInterrupt:
      client.maybe_kill()
      return executor.SIGINT_CODE
    except executor.UnreachableError as e:
      print(
        f'Worker {e.worker} was unrechable, run `rx init` to get a new '
        'instance.')
      return e.code


class VersionCommand:
  """Prints the version."""

  def run(self) -> int:
    print(local.VERSION)
    return 0


def main(argv):
  logging.get_absl_handler().python_handler.use_absl_log_file(
    program_name='rx', log_dir=tempfile.gettempdir())
  if len(argv) == 1:
    print('No command given.\n')
    app.usage(shorthelp=True)
    return -1
  cmd_to_run = argv[1]
  try:
    if cmd_to_run == 'init':
      cmd = init.InitCommand()
    elif cmd_to_run == 'version':
      cmd = VersionCommand()
    else:
      if cmd_to_run == 'run':
        # "rx run foo" is generally the same as "rx foo", but the extra "run"
        # can be handy when running a script called "init", say, on the remote
        # machine.
        argv = argv[1:]
      cmd = ExecCommand(argv[1:])
    return cmd.run()
  except KeyboardInterrupt:
    return executor.SIGINT_CODE


def run():
  app.run(main)



import pathlib
import sys

from absl import flags
from absl import logging

from rx.client import grpc_helper
from rx.client import init_client
from rx.client.configuration import config_base
from rx.client.configuration import local

_DRY_RUN = flags.DEFINE_bool(
  'dry-run', False, 'Shows a list of files that would be uploaded by rx init')


class InitCommand:
  """Initialize (or reinitialize) the remote."""

  def __init__(self):
    self._cwd = (
      pathlib.Path(config_base.RX_ROOT.value) if config_base.RX_ROOT.value else
      pathlib.Path.cwd())
    self._config = local.find_local_config(self._cwd)

  def run(self) -> int:
    if self._config is None:
      local.install_local_files(self._cwd)
    else:
      logging.info('Workspace already exists, resetting it.')
    try:
      self._config = local.create_local_config(self._cwd)
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        client = init_client.Client(ch, self._config)
        if _DRY_RUN.value:
          client.dry_run()
          return 0
        client.create_user_or_log_in()
        return client.init()
    except init_client.InitError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code

if __name__ == '__main__':
  print('Call exec.')
